import compose.config.config as compose

import os
import yaml

import os.path

from collections import namedtuple


def parse(working_dir, f):
    with open(os.path.join(working_dir, f), 'r') as config_file:
        return Config(yaml.safe_load(config_file), working_dir, f)


class NoSuchTargetException(Exception):
    def __init__(self, target):
        self.target = target


class Config(object):
    def __init__(self, yml, working_dir=os.getcwd(), filename='construi.yml'):
        self.yml = yml
        self.working_dir = working_dir
        self.filename = filename

    def __getattr__(self, name):
        return self.yml[name]

    @property
    def project_name(self):
        return os.path.basename(self.working_dir)

    def for_target(self, target):
        config_files = [
            self.base_config(target), self.target_config(target),
            self.workspace_config(target)
        ]

        config_details = compose.ConfigDetails(self.working_dir, config_files)

        target_yml = self.target_yml(target)

        construi = {
            'before': target_yml['before'] if 'before' in target_yml else [],
            'name': target,
            'run': self.target_yml(target).get('run', []),
            'shell': self.target_yml(target).get('shell', None),
            'project_name': self.project_name
        }

        return TargetConfig(construi, compose.load(config_details))

    def base_config(self, target):
        base_yml = dict(self.yml)

        delete(base_yml, 'default', 'targets')

        # If the target specified build and/or image remove it from base.
        # IMO this behavior is more intuitive for construi than the default
        # merging behavior of the compose v2 schema.
        target_yml = self.target_yml(target)

        if 'build' in target_yml or 'image' in target_yml:
            delete(base_yml, 'build', 'image')

        return self.create_config_file({target: base_yml})

    def target_config(self, target):
        target_yml = self.target_yml(target)

        services = self.get_links(target_yml)

        delete(target_yml, 'before', 'links', 'run', 'shell')

        target_yml['links'] = [s for s in services.keys()]

        services[target] = target_yml

        return self.create_config_file(services)

    def target_yml(self, target):
        try:
            yml = self.yml['targets'][target]
        except KeyError:
            raise NoSuchTargetException(target)

        if type(yml) is str:
            yml = {'run': [yml]}
        if 'run' in yml and type(yml['run']) is str:
            yml['run'] = [yml['run']]

        return yml.copy()

    def get_links(self, yml):
        links = {}

        if 'links' in yml:
            links = yml['links']

        return links

    def workspace_config(self, name):
        config = {
            'working_dir': self.working_dir,
            'volumes': ["%s:%s" % (self.working_dir, self.working_dir)]
        }

        return self.create_config_file({name: config})

    def create_config_file(self, yml):
        return compose.ConfigFile(self.filename,
                                  {'version': '2',
                                   'services': yml})


class TargetConfig(namedtuple('_TargetConfig', 'construi compose')):
    @property
    def services(self):
        return self.compose.services


def delete(hsh, *keys):
    for key in keys:
        if key in hsh:
            del hsh[key]



from construi.config import Config, NoSuchTargetException, TargetConfig

from compose.config.types import VolumeSpec

import os
import pytest
import sys
import yaml


class TestConfig(object):
    @property
    def working_dir(self):
        return os.path.dirname(os.path.realpath(__file__))

    def config(self, yml, target):
        return Config(yml, self.working_dir).for_target(target)

    def test_for_target(self):
        yml = yaml.load("""
          image: java:latest

          targets:
            build: mvn install
        """)

        config = self.config(yml, 'build')

        assert config.construi['before'] == []
        assert config.construi['name'] == 'build'
        assert config.construi['run'] == ['mvn install']

        assert len(config.services) == 1

        service = config.services[0]
        assert service['working_dir'] == self.working_dir
        assert service['name'] == 'build'
        assert service['image'] == 'java:latest'
        assert service['volumes'] == [
            VolumeSpec(
                external=self.working_dir,
                internal=self.working_dir,
                mode='rw')
        ]

    def test_no_such_target(self):
        yml = yaml.load("""
          image: java:latest

          targets:
            build: mvn install
        """)

        with pytest.raises(NoSuchTargetException):
            config = self.config(yml, 'does_not_exist')

    def test_build_and_image_handling(self):
        yml = yaml.load("""
          image: java:latest

          targets:
            target:
              build: Dockerfile.dummy
        """)

        config = self.config(yml, 'target')

        assert len(config.services) == 1

        service = config.services[0]
        assert 'image' not in service
        assert service['build'] == {
            'context': os.path.join(self.working_dir, 'Dockerfile.dummy')
        }

    def test_no_run(self):
        yml = yaml.load("""
          image: java:latest

          targets:
            build:
              before:
                - a
                - b
        """)

        config = self.config(yml, 'build')

        assert config.construi['before'] == ['a', 'b']
        assert config.construi['run'] == []

    def test_volumes(self):
        yml = yaml.load("""
          image: java:latest
          volumes:
            - /a:/a

          targets:
            build:
              volumes:
                - /b:/b
              run: run.sh
        """)

        config = self.config(yml, 'build')

        assert config.construi['run'] == ['run.sh']
        assert len(config.services) == 1
        assert VolumeSpec(
            external='/b', internal='/b',
            mode='rw') in config.services[0]['volumes']
        assert VolumeSpec(
            external='/a', internal='/a',
            mode='rw') in config.services[0]['volumes']

    def test_links(self):
        yml = yaml.load("""
            image: java:latest

            targets:
              build:
                run: run.sh
                links:
                  mysql:
                    image: mysql:latest
        """)

        config = self.config(yml, 'build')

        print(config)

        assert config.construi['run'] == ['run.sh']
        assert len(config.services) == 2

        build = get_service(config, 'build')
        assert build['links'] == ['mysql']
        assert build['image'] == 'java:latest'

        mysql = get_service(config, 'mysql')
        assert mysql['image'] == 'mysql:latest'


def get_service(target_config, name):
    return next(s for s in target_config.services if s['name'] == name)



from .config import parse, NoSuchTargetException
from .target import BuildFailedException, RunContext, Target
from .__version__ import __version__

from argparse import ArgumentParser

from compose.errors import OperationFailedError

import construi.console as console

import logging
import os
import sys
import traceback

try:
    from shlex import quote
except ImportError:
    from pipes import quote


def main():
    setup_logging()

    args = parse_args()

    config = parse(args.basedir, 'construi.yml')

    if args.list_targets:
        list_targets(config)
        sys.exit(0)

    target = args.target or config.default

    os.environ['CONSTRUI_ARGS'] = ' '.join(
        [quote(a) for a in args.construi_args])

    try:
        Target(config.for_target(target)).invoke(
            RunContext(config, args.dry_run))
    except BuildFailedException:
        console.error("\nBuild Failed.\n")
        sys.exit(1)
    except NoSuchTargetException, e:
        console.error("\nNo such target: {}\n".format(e.target))
        sys.exit(1)
    except OperationFailedError, e:
        console.error("\nUnexpected Error: {}\n".format(e.msg))
        traceback.print_exc()
        sys.exit(1)
    except KeyboardInterrupt:
        console.warn("\nBuild Interrupted.")
        sys.exit(1)

    console.info("\nBuild Succeeded.\n")


def setup_logging():
    root_logger = logging.getLogger()
    root_logger.addHandler(logging.StreamHandler(sys.stdout))
    root_logger.setLevel(logging.INFO)

    logging.getLogger("requests").propagate = False


def parse_args():
    parser = ArgumentParser(prog='construi', description='Run construi')

    parser.add_argument('--basedir', metavar='DIR', default=os.getcwd())
    parser.add_argument('-n', '--dry-run', action='store_true')
    parser.add_argument('-T', '--list-targets', action='store_true')
    parser.add_argument(
        '-v', '--version', action='version', version=__version__)

    parser.add_argument('target', metavar='TARGET', nargs='?')
    parser.add_argument('construi_args', metavar='CONSTRUI_ARGS', nargs='*')

    return parser.parse_args()


def list_targets(config):
    targets = config.targets.keys()

    targets.sort()

    for target in targets:
        print(target)



from compose.cli import colors


def error(msg):
    print(colors.red(msg))


def warn(msg):
    print(colors.yellow(msg))


def info(msg):
    print(colors.green(msg))


def progress(msg):
    info("\n%s" % msg)



import construi.console as console

from compose.project import Project
from compose.cli.docker_client import docker_client
from compose.service import ConvergenceStrategy

import dockerpty
import sys
import os
import os.path
import shlex


class BuildFailedException(Exception):
    def __init__(self):
        pass


class Target(object):
    def __init__(self, config):
        self.config = config
        self.project = Project.from_config(
            "construi_%s" % self.config.construi['project_name'],
            config.compose, docker_client(os.environ))

    @property
    def before(self):
        return self.config.construi['before']

    @property
    def client(self):
        return self.project.client

    @property
    def commands(self):
        return self.config.construi.get('run', [])

    @property
    def name(self):
        return self.config.construi['name']

    @property
    def shell(self):
        return self.config.construi['shell']

    @property
    def service(self):
        return self.project.get_service(self.name)

    @property
    def linked_services(self):
        return [s for s in self.project.service_names if s != self.name]

    def invoke(self, run_ctx):
        console.progress("** Invoke %s" % self.name)

        for target in self.before:
            Target(run_ctx.config.for_target(target)).invoke(run_ctx)

        if run_ctx.is_executed(self.name):
            console.progress("** Skipped %s" % self.name)
            return

        if self.commands:
            dry_run = '(dry run)' if run_ctx.dry_run else ''
            console.progress("** Execute %s %s" % (self.name, dry_run))

            if not run_ctx.dry_run:
                self.run()

        run_ctx.mark_executed(self.name)

    def run(self):
        try:
            self.setup()

            self.start_linked_services()

            for command in self.commands:
                self.run_command(command)

            console.progress('Done.')

        finally:
            self.cleanup()

    def run_command(self, command):
        if self.shell:
            to_run = shlex.split(self.shell) + [command]
            console.progress("(%s)> %s" % (self.shell, command))
        else:
            to_run = command
            console.progress("> %s" % command)

        container = self.service.create_container(
            one_off=True,
            command=to_run,
            tty=False,
            stdin_open=True,
            detach=False)

        try:
            dockerpty.start(self.client, container.id, interactive=False)

            if container.wait() != 0:
                raise BuildFailedException()
                sys.exit(1)

        finally:
            self.client.remove_container(container.id, force=True)

    def setup(self):
        console.progress('Building Images...')
        self.project.build()

        console.progress('Pulling Images...')
        self.project.pull()

        self.project.initialize()

    def start_linked_services(self):
        if self.linked_services:
            self.project.up(
                service_names=self.linked_services,
                start_deps=True,
                strategy=ConvergenceStrategy.always)

    def cleanup(self):
        console.progress('Cleaning up...')
        self.project.kill()
        self.project.remove_stopped(None, v=True)
        self.project.networks.remove()


class RunContext(object):
    def __init__(self, config, dry_run=False):
        self.config = config
        self.dry_run = dry_run
        self.executed = set()

    def mark_executed(self, target):
        self.executed.add(target)

    def is_executed(self, target):
        return target in self.executed



from setuptools import find_packages, setup

import codecs
import distutils
import os
import re


def read(*parts):
    path = os.path.join(os.path.dirname(__file__), *parts)
    with codecs.open(path, encoding='utf-8') as fobj:
        return fobj.read()


def find_version():
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              read('construi', '__version__.py'), re.M)

    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


class YapfCommand(distutils.cmd.Command):
    description = 'Format python files using yapf'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import yapf
        yapf.main(['yapf', '--recursive', '--in-place', 'setup.py', 'test'] +
                  find_packages())


requires = {
    'install': ['PyYAML == 3.11', 'docker-compose == 1.11.2', 'six == 1.10.0'],
    'setup': ['flake8 == 3.0.4', 'pytest-runner == 2.6.2', 'yapf == 0.16.1'],
    'tests': ['pytest == 2.8.5', 'pytest-cov == 2.2.0']
}

summary = 'Use Docker to define your build environment'

setup(
    name='construi',
    version=find_version(),
    url='https://github.com/lstephen/construi',
    license='Apache License 2.0',
    description=summary,
    long_description=summary,
    author='Levi Stephen',
    author_email='levi.stephen@gmail.com',
    zip_safe=True,
    packages=find_packages(),
    install_requires=requires['install'],
    setup_requires=requires['setup'],
    tests_require=requires['tests'],
    entry_points={'console_scripts': [
        'construi=construi.cli:main',
    ]},
    cmdclass={'yapf': YapfCommand},
    classifiers=[
        'Development Status :: 4 - Beta', 'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4'
    ])
